package sun.nio.fs;

import java.nio.file.attribute.FileAttribute;
import java.util.Set;
import java.nio.file.attribute.AclEntryPermission;
import java.util.EnumSet;
import java.nio.file.attribute.AclEntryFlag;
import java.nio.file.attribute.AclEntryType;
import java.nio.file.attribute.UserPrincipal;
import java.util.Iterator;
import java.nio.file.ProviderMismatchException;
import java.io.IOException;
import java.util.ArrayList;
import java.nio.file.attribute.AclEntry;
import java.util.List;
import sun.misc.Unsafe;

class WindowsSecurityDescriptor
{
    private static final Unsafe unsafe;
    private static final short SIZEOF_ACL = 8;
    private static final short SIZEOF_ACCESS_ALLOWED_ACE = 12;
    private static final short SIZEOF_ACCESS_DENIED_ACE = 12;
    private static final short SIZEOF_SECURITY_DESCRIPTOR = 20;
    private static final short OFFSETOF_TYPE = 0;
    private static final short OFFSETOF_FLAGS = 1;
    private static final short OFFSETOF_ACCESS_MASK = 4;
    private static final short OFFSETOF_SID = 8;
    private static final WindowsSecurityDescriptor NULL_DESCRIPTOR;
    private final List<Long> sidList;
    private final NativeBuffer aclBuffer;
    private final NativeBuffer sdBuffer;
    
    private WindowsSecurityDescriptor() {
        this.sidList = null;
        this.aclBuffer = null;
        this.sdBuffer = null;
    }
    
    private WindowsSecurityDescriptor(List<AclEntry> c) throws IOException {
        boolean n = false;
        ArrayList list = new ArrayList(c);
        this.sidList = new ArrayList(list.size());
        try {
            int n2 = 8;
            Iterator<Object> iterator = list.iterator();
            while (iterator.hasNext()) {
                UserPrincipal principal = ((AclEntry)iterator.next()).principal();
                if (!(principal instanceof WindowsUserPrincipals.User)) {
                    throw new ProviderMismatchException();
                }
                String sidString = ((WindowsUserPrincipals.User)principal).sidString();
                try {
                    long convertStringSidToSid = WindowsNativeDispatcher.ConvertStringSidToSid(sidString);
                    this.sidList.add(Long.valueOf(convertStringSidToSid));
                    n2 += WindowsNativeDispatcher.GetLengthSid(convertStringSidToSid) + Math.max(12, 12);
                }
                catch (WindowsException ex) {
                    throw new IOException("Failed to get SID for " + principal.getName() + ": " + ex.errorString());
                }
            }
            this.aclBuffer = NativeBuffers.getNativeBuffer(n2);
            this.sdBuffer = NativeBuffers.getNativeBuffer(20);
            WindowsNativeDispatcher.InitializeAcl(this.aclBuffer.address(), n2);
            for (int i = 0; i < list.size(); ++i) {
                AclEntry aclEntry = (AclEntry)list.get(i);
                long longValue = (long)this.sidList.get(i);
                try {
                    encode(aclEntry, longValue, this.aclBuffer.address());
                }
                catch (WindowsException ex2) {
                    throw new IOException("Failed to encode ACE: " + ex2.errorString());
                }
            }
            WindowsNativeDispatcher.InitializeSecurityDescriptor(this.sdBuffer.address());
            WindowsNativeDispatcher.SetSecurityDescriptorDacl(this.sdBuffer.address(), this.aclBuffer.address());
            n = true;
        }
        catch (WindowsException ex3) {
            throw new IOException(ex3.getMessage());
        }
        finally {
            if (n == 0) {
                this.release();
            }
        }
    }
    
    void release() {
        if (this.sdBuffer != null) {
            this.sdBuffer.release();
        }
        if (this.aclBuffer != null) {
            this.aclBuffer.release();
        }
        if (this.sidList != null) {
            Iterator<Long> iterator = this.sidList.iterator();
            while (iterator.hasNext()) {
                WindowsNativeDispatcher.LocalFree((long)iterator.next());
            }
        }
    }
    
    long address() {
        return (this.sdBuffer == null) ? 0L : this.sdBuffer.address();
    }
    
    private static AclEntry decode(long n) throws IOException {
        int byte1 = WindowsSecurityDescriptor.unsafe.getByte(n + 0L);
        if (byte1 != 0 && byte1 != 1) {
            return null;
        }
        AclEntryType type;
        if (byte1 == 0) {
            type = AclEntryType.ALLOW;
        }
        else {
            type = AclEntryType.DENY;
        }
        int byte2 = WindowsSecurityDescriptor.unsafe.getByte(n + 1L);
        EnumSet<AclEntryFlag> none = EnumSet.noneOf(AclEntryFlag.class);
        if ((byte2 & 0x1) != 0x0) {
            none.add(AclEntryFlag.FILE_INHERIT);
        }
        if ((byte2 & 0x2) != 0x0) {
            none.add(AclEntryFlag.DIRECTORY_INHERIT);
        }
        if ((byte2 & 0x4) != 0x0) {
            none.add(AclEntryFlag.NO_PROPAGATE_INHERIT);
        }
        if ((byte2 & 0x8) != 0x0) {
            none.add(AclEntryFlag.INHERIT_ONLY);
        }
        int int1 = WindowsSecurityDescriptor.unsafe.getInt(n + 4L);
        EnumSet<AclEntryPermission> none2 = EnumSet.noneOf(AclEntryPermission.class);
        if ((int1 & 0x1) > 0) {
            none2.add(AclEntryPermission.READ_DATA);
        }
        if ((int1 & 0x2) > 0) {
            none2.add(AclEntryPermission.WRITE_DATA);
        }
        if ((int1 & 0x4) > 0) {
            none2.add(AclEntryPermission.APPEND_DATA);
        }
        if ((int1 & 0x8) > 0) {
            none2.add(AclEntryPermission.READ_NAMED_ATTRS);
        }
        if ((int1 & 0x10) > 0) {
            none2.add(AclEntryPermission.WRITE_NAMED_ATTRS);
        }
        if ((int1 & 0x20) > 0) {
            none2.add(AclEntryPermission.EXECUTE);
        }
        if ((int1 & 0x40) > 0) {
            none2.add(AclEntryPermission.DELETE_CHILD);
        }
        if ((int1 & 0x80) > 0) {
            none2.add(AclEntryPermission.READ_ATTRIBUTES);
        }
        if ((int1 & 0x100) > 0) {
            none2.add(AclEntryPermission.WRITE_ATTRIBUTES);
        }
        if ((int1 & 0x10000) > 0) {
            none2.add(AclEntryPermission.DELETE);
        }
        if ((int1 & 0x20000) > 0) {
            none2.add(AclEntryPermission.READ_ACL);
        }
        if ((int1 & 0x40000) > 0) {
            none2.add(AclEntryPermission.WRITE_ACL);
        }
        if ((int1 & 0x80000) > 0) {
            none2.add(AclEntryPermission.WRITE_OWNER);
        }
        if ((int1 & 0x100000) > 0) {
            none2.add(AclEntryPermission.SYNCHRONIZE);
        }
        return AclEntry.newBuilder().setType(type).setPrincipal(WindowsUserPrincipals.fromSid(n + 8L)).setFlags(none).setPermissions(none2).build();
    }
    
    private static void encode(AclEntry aclEntry, long n, long n2) throws WindowsException {
        if (aclEntry.type() != AclEntryType.ALLOW && aclEntry.type() != AclEntryType.DENY) {
            return;
        }
        int n3 = aclEntry.type() == AclEntryType.ALLOW;
        Set<AclEntryPermission> permissions = aclEntry.permissions();
        int n4 = 0;
        if (permissions.contains(AclEntryPermission.READ_DATA)) {
            n4 |= 0x1;
        }
        if (permissions.contains(AclEntryPermission.WRITE_DATA)) {
            n4 |= 0x2;
        }
        if (permissions.contains(AclEntryPermission.APPEND_DATA)) {
            n4 |= 0x4;
        }
        if (permissions.contains(AclEntryPermission.READ_NAMED_ATTRS)) {
            n4 |= 0x8;
        }
        if (permissions.contains(AclEntryPermission.WRITE_NAMED_ATTRS)) {
            n4 |= 0x10;
        }
        if (permissions.contains(AclEntryPermission.EXECUTE)) {
            n4 |= 0x20;
        }
        if (permissions.contains(AclEntryPermission.DELETE_CHILD)) {
            n4 |= 0x40;
        }
        if (permissions.contains(AclEntryPermission.READ_ATTRIBUTES)) {
            n4 |= 0x80;
        }
        if (permissions.contains(AclEntryPermission.WRITE_ATTRIBUTES)) {
            n4 |= 0x100;
        }
        if (permissions.contains(AclEntryPermission.DELETE)) {
            n4 |= 0x10000;
        }
        if (permissions.contains(AclEntryPermission.READ_ACL)) {
            n4 |= 0x20000;
        }
        if (permissions.contains(AclEntryPermission.WRITE_ACL)) {
            n4 |= 0x40000;
        }
        if (permissions.contains(AclEntryPermission.WRITE_OWNER)) {
            n4 |= 0x80000;
        }
        if (permissions.contains(AclEntryPermission.SYNCHRONIZE)) {
            n4 |= 0x100000;
        }
        Set<AclEntryFlag> flags = aclEntry.flags();
        int n5 = 0;
        if (flags.contains(AclEntryFlag.FILE_INHERIT)) {
            n5 = (byte)(n5 | 0x1);
        }
        if (flags.contains(AclEntryFlag.DIRECTORY_INHERIT)) {
            n5 = (byte)(n5 | 0x2);
        }
        if (flags.contains(AclEntryFlag.NO_PROPAGATE_INHERIT)) {
            n5 = (byte)(n5 | 0x4);
        }
        if (flags.contains(AclEntryFlag.INHERIT_ONLY)) {
            n5 = (byte)(n5 | 0x8);
        }
        if (!(n3 == 0)) {
            WindowsNativeDispatcher.AddAccessAllowedAceEx(n2, n5, n4, n);
        }
        else {
            WindowsNativeDispatcher.AddAccessDeniedAceEx(n2, n5, n4, n);
        }
        return;
    }
    
    static WindowsSecurityDescriptor create(List<AclEntry> list) throws IOException {
        return new WindowsSecurityDescriptor(list);
    }
    
    static WindowsSecurityDescriptor fromAttribute(FileAttribute<?>... array) throws IOException {
        WindowsSecurityDescriptor null_DESCRIPTOR = WindowsSecurityDescriptor.NULL_DESCRIPTOR;
        int length = array.length;
        int i = 0;
        while (i < length) {
            FileAttribute<?> fileAttribute = array[i];
            if (null_DESCRIPTOR != WindowsSecurityDescriptor.NULL_DESCRIPTOR) {
                null_DESCRIPTOR.release();
            }
            if (fileAttribute == null) {
                throw new NullPointerException();
            }
            if (fileAttribute.name().equals("acl:acl")) {
                null_DESCRIPTOR = new WindowsSecurityDescriptor((List)fileAttribute.value());
                ++i;
            }
            else {
                throw new UnsupportedOperationException("'" + fileAttribute.name() + "' not supported as initial attribute");
            }
        }
        return null_DESCRIPTOR;
    }
    
    static List<AclEntry> getAcl(long n) throws IOException {
        long getSecurityDescriptorDacl = WindowsNativeDispatcher.GetSecurityDescriptorDacl(n);
        int aceCount;
        if (getSecurityDescriptorDacl == 0L) {
            aceCount = 0;
        }
        else {
            aceCount = WindowsNativeDispatcher.GetAclInformation(getSecurityDescriptorDacl).aceCount();
        }
        ArrayList o = new ArrayList(aceCount);
        for (int i = 0; i < aceCount; ++i) {
            AclEntry decode = decode(WindowsNativeDispatcher.GetAce(getSecurityDescriptorDacl, i));
            if (decode != null) {
                o.add(decode);
            }
        }
        return o;
    }
    
    static {
        unsafe = Unsafe.getUnsafe();
        NULL_DESCRIPTOR = new WindowsSecurityDescriptor();
    }
}
